==> cstream-1.0.tar.gz: cstream-1.0/cstream.c <==
/*
 * Copyright 1998, 1999 Martin Cracauer
 *
 * See the file COPYRIGHT in the distribution package for copying terms.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <errno.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>

volatile sig_atomic_t signal_shutdown;
volatile sig_atomic_t signal_report;

struct options {
  int v;
  int b;
  long long n;
  int t;
  char *i;
  char *o;
  char I;
  char O;
  char *p;
};

struct progstate {
  int ifd;
  int ofd;
  long long bytes_transferred;
};

static void sighup(int signal)
{
  signal_shutdown = 1;
}

static void sigusr1(int signal)
{
  signal_report = 1;
}

static struct options *default_options(struct options *const o)
{
  bzero(o, sizeof(struct options));
  return o;
}

static void init(struct options *const o, struct progstate *const state)
{
  if (o->b == 0)
    o->b = 8192;

  if (o->i == NULL || o->i[0] == '\0')
    state->ifd = 0;
  else if (!strcmp(o->i, "-"))
    state->ifd = -1;
  else  {
    if (o->I == 'f') {
      unlink(o->i);
      if (mkfifo(o->i, 0666) == -1) {
	perror("mkfifo in");
	exit(2);
      }
    }
    /* not else */
    state->ifd = open(o->i, O_RDWR);
    if (state->ifd == -1) {
      perror("open input file\n");
      exit(2);
    }
  }

  if (o->o == NULL || o->o[0] == '\0')
    state->ofd = 1;
  else if (!strcmp(o->o, "-"))
    state->ofd = -1;
  else {
    if (o->O == 'f') {
      unlink(o->o);
      if (mkfifo(o->o, 0666) == -1) {
	perror("mkfifo out");
	exit(2);
      }
    }
    /* not else */
    state->ofd = open(o->o, O_RDWR | O_CREAT | O_TRUNC);
    if (state->ofd == -1) {
      perror("open output file");
      exit(2);
    }
    if (o->v >= 2)
      fprintf(stderr,"Files open at %d/%d\n", state->ifd, state->ofd);
  }
  if (o->p) {
    FILE *f;

    f = fopen(o->p, "w");
    if (f == NULL) {
      perror("fopen/write/pidfile");
      exit(2);
    }
    fprintf(f, "%d\n", getpid());
    fclose(f);
  }
}

static void report(const struct progstate *const state, 
		   const double starttime)
{
  struct timeval t2;
  double sofar;
  double rate;
    
  if (gettimeofday(&t2, NULL) == -1) {
    perror("gettimeofday");
    exit(2);
  }
  sofar = ((double)t2.tv_sec + (double)t2.tv_usec / 1000000.0) -
    starttime;
  rate = (double)state->bytes_transferred / sofar;
  fprintf(stderr,"%.0f bytes %.2f seconds %.0f bytes/sec",
	  (double)state->bytes_transferred,
	  sofar,
	  (double)state->bytes_transferred / sofar);
  if (rate >= 1000000)
    fprintf(stderr, " %.2f MB/sec", rate/1024.0/1024.0);
  else
    if (rate >= 1000)
      fprintf(stderr, " %.2f KB/sec", rate/1024.0);
  fprintf(stderr, "\n");
}


static void loop(struct options *const o, struct progstate *const state)
{
  char *buf;
  int ret = 0; /* Kill warning */
  struct timeval t;
  double starttime = 0.0; /* Kill warning */

  buf = malloc(o->b);
  if (buf == NULL) {
    perror("malloc");
    exit(2);
  }

  state->bytes_transferred = 0;
  if (o->t || o->v >= 1) {
    if (gettimeofday(&t, NULL) == -1) {
      perror("gettimeofday");
      exit(2);
    }
    starttime = (double)t.tv_sec + (double)t.tv_usec / 1000000.0;
  }

  signal_shutdown = 0;
  signal_report = 0;
  signal(SIGHUP, sighup);
  signal(SIGUSR1, sigusr1);
  while(signal_shutdown == 0) {

    /* 
     * Sleep long enough to satisfy througput limit.
     * Do this before the read, so that sending process will be blocked
     * exactly the right amount of time.
     *
     */

    if (state->bytes_transferred && o->t) {
      struct timeval t2;
      double sofar;
      double theory;

      if (gettimeofday(&t2, NULL) == -1) {
	perror("gettimeofday");
	exit(2);
      }
      sofar = ((double)t2.tv_sec + (double)t2.tv_usec / 1000000.0) -
	starttime;

      /* In theory we should have needed how many useconds? */
      theory = (double)state->bytes_transferred / (double)o->t;
      if (o->v >= 2)
	fprintf(stderr,"We needed %g seconds, should be %g\n", sofar, theory);
      if (sofar < theory)
	usleep((theory - sofar) * 1000000.0);
    }

    if (state->ifd != -1) {
      if (o->v >= 2)
	fprintf(stderr,"Trying to read %d bytes\n", o->b);
      if (o->n && state->bytes_transferred + o->b > o->n)
	ret = read(state->ifd, buf, o->n - state->bytes_transferred);
      else
	ret = read(state->ifd, buf, o->b);
      if (o->v >= 2)
	fprintf(stderr,"read %d bytes\n", ret);
      if (ret == 0)
	break;
      if (ret == -1) {
	if (errno == EINTR)
	  ret = 0; /* 
		    * Not a mistake, just a hack. write should have 0
		    * byte to write.  I asked for 0 as indicator for
		    * end-of-file above.  Do not put the ret=0 ->
		    * break line after this!
		    * 
		    */
	else {
	  perror("read");
	  exit(2);
	}
      }
    }
    else {
      /* Just generate data */
      if (o->n && state->bytes_transferred + (long long)o->b > o->n)
	ret = o->n - state->bytes_transferred;
      else
	ret = o->b;
    }

    if (state->ofd != -1) {
      ret = write(state->ofd, buf, ret);
      if (ret == -1) {
	perror("write");
	exit(2);
      }
      if (o->v >= 2)
	fprintf(stderr,"write %d bytes\n", ret);
    }
    state->bytes_transferred += ret;
    if (o->n && state->bytes_transferred >= o->n)
      break;

    if (signal_report) {
      report(state, starttime);
      signal_report = 0;
    }
  }
  if (o->v >= 1)
    report(state, starttime);
  free(buf);
}

void shutdown(const struct options *const o, struct progstate *const state)
{

  if (o->p)
    if (unlink(o->p) == -1)
      perror("unlink pidfile - continuing");

  if (state->ofd != 1 && state->ofd != 2 && state->ofd != -1)
    if (close(state->ofd) == -1) {
      perror("close out");
    }
  if (state->ifd != 0 && state->ifd != -1)
    if (close(state->ifd) == -1) {
      perror("close in");
    }
}

static void usage_message(void)
{
  fprintf(stderr, "cstream by Martin Cracauer - version " VERSION "\n");
  fprintf(stderr,
	  "-v <n> = verbose [default: off]\n"
	  "         0 = nothing\n"
	  "         1 = report bytes transferred and throughput\n"
	  "         3 = verbose stats on every read/write\n"
	  "-b <n> = blocksize [default: 8192]\n"
	  "-n <n> = overall size of data [default: unlimited]\n"
	  "-t <n> = throughput in bytes/sec [default: unlimited]\n"
	  "-i <s> = name of input file, - = generate stream yourself\n"
	  "         to use stdin, use -i ''\n"
	  "-o <s> = name of output file, - = just sink data\n"
	  "         to use stdout, -o ''\n"
	  "-I <c> = Type of input file, 'f' = fifo (create it)\n"
	  "-O <c> = Type of output file, 'f' = fifo (create it)\n"
	  "-p <s> = Write pid as ascii interger to file named <s>\n"
	  "SIGHUP causes loop end after next buffer transfer\n"
	  "SIGUSR1 causes statistics to be written to stderr\n"
	  );
}

static void usage(void)
{
  usage_message();
  exit(1);
}


int main(int argc, char *const argv[])
{
  struct options o;
  struct progstate state;
  int ch;

  default_options(&o);

  while ((ch = getopt(argc, argv, "?v:b:n:t:i:o:I:O:p:")) != -1) {
    switch(ch) {
    case '?': usage_message(); exit(0);
    case 'v': {
      if (optarg[1] != '\0' ||
	  optarg[0] != '0' ||
	  optarg[1] != '0' ||
	  optarg[2] != '0' ||
	  optarg[3] != '0') {
	fprintf(stderr, "Error: option -v requires a decimal numeric argument "
		"between 0 and 3.\nType -? for usage instructions\n");
	exit(1);
      }
      o.v = atoi(optarg); 
      break;
    }
    case 'b': o.b = atoi(optarg); break;
    case 'n': o.n = atoi(optarg);
      if (tolower(optarg[strlen(optarg)-1]) == 'k')
	o.n *= 1024;
      else if (tolower(optarg[strlen(optarg)-1]) == 'm')
	o.n *= 1024 * 1024;
      else if (tolower(optarg[strlen(optarg)-1]) == 'g')
	o.n *= 1024 * 1024 * 1024;
      break;
    case 't': o.t = atoi(optarg); break;
    case 'i': o.i = strdup(optarg); break;
    case 'o': o.o = strdup(optarg); break;
    case 'I': o.I = optarg[0]; break;
    case 'O': o.O = optarg[0]; break;
    case 'p': o.p = strdup(optarg); break;
    default: usage();
    }
  }
  argc -= optind;
  argv += optind;
  if (argc != 0)
    usage();

  init(&o, &state);

  loop(&o, &state);

  shutdown(&o, &state);

  /* Freeing of memory omitted */

  return 0;
}

==> cstream-1.1.tar.gz: cstream-1.1/cstream.c <==
/*
 * Copyright 1998, 1999 Martin Cracauer
 *
 * See the file COPYRIGHT in the distribution package for copying terms.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <errno.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>

volatile sig_atomic_t signal_shutdown;
volatile sig_atomic_t signal_report;

struct options {
  int v;
  int b;
  long long n;
  int t;
  char *i;
  char *o;
  char I;
  char O;
  char *p;
};

struct progstate {
  int ifd;
  int ofd;
  long long bytes_transferred;
};

static void sighup(int signal)
{
  signal_shutdown = 1;
}

static void sigusr1(int signal)
{
  signal_report = 1;
}

static struct options *default_options(struct options *const o)
{
  bzero(o, sizeof(struct options));
  return o;
}

static void init(struct options *const o, struct progstate *const state)
{
  if (o->b == 0)
    o->b = 8192;

  if (o->i == NULL || o->i[0] == '\0')
    state->ifd = 0;
  else if (!strcmp(o->i, "-"))
    state->ifd = -1;
  else  {
    if (o->I == 'f') {
      unlink(o->i);
      if (mkfifo(o->i, 0666) == -1) {
	perror("mkfifo in");
	exit(2);
      }
    }
    /* not else */
    state->ifd = open(o->i, O_RDWR);
    if (state->ifd == -1) {
      perror("open input file\n");
      exit(2);
    }
  }

  if (o->o == NULL || o->o[0] == '\0')
    state->ofd = 1;
  else if (!strcmp(o->o, "-"))
    state->ofd = -1;
  else {
    if (o->O == 'f') {
      unlink(o->o);
      if (mkfifo(o->o, 0666) == -1) {
	perror("mkfifo out");
	exit(2);
      }
    }
    /* not else */
    state->ofd = open(o->o, O_RDWR | O_CREAT | O_TRUNC);
    if (state->ofd == -1) {
      perror("open output file");
      exit(2);
    }
    if (o->v >= 2)
      fprintf(stderr,"Files open at %d/%d\n", state->ifd, state->ofd);
  }
  if (o->p) {
    FILE *f;

    f = fopen(o->p, "w");
    if (f == NULL) {
      perror("fopen/write/pidfile");
      exit(2);
    }
    fprintf(f, "%d\n", getpid());
    fclose(f);
  }
}

static void report(const struct progstate *const state, 
		   const double starttime)
{
  struct timeval t2;
  double sofar;
  double rate;
    
  if (gettimeofday(&t2, NULL) == -1) {
    perror("gettimeofday");
    exit(2);
  }
  sofar = ((double)t2.tv_sec + (double)t2.tv_usec / 1000000.0) -
    starttime;
  rate = (double)state->bytes_transferred / sofar;
  fprintf(stderr,"%.0f bytes %.2f seconds %.0f bytes/sec",
	  (double)state->bytes_transferred,
	  sofar,
	  (double)state->bytes_transferred / sofar);
  if (rate >= 1000000)
    fprintf(stderr, " %.2f MB/sec", rate/1024.0/1024.0);
  else
    if (rate >= 1000)
      fprintf(stderr, " %.2f KB/sec", rate/1024.0);
  fprintf(stderr, "\n");
}


static void loop(struct options *const o, struct progstate *const state)
{
  char *buf;
  int ret = 0; /* Kill warning */
  struct timeval t;
  double starttime = 0.0; /* Kill warning */

  buf = malloc(o->b);
  if (buf == NULL) {
    perror("malloc");
    exit(2);
  }

  state->bytes_transferred = 0;
  if (o->t || o->v >= 1) {
    if (gettimeofday(&t, NULL) == -1) {
      perror("gettimeofday");
      exit(2);
    }
    starttime = (double)t.tv_sec + (double)t.tv_usec / 1000000.0;
  }

  signal_shutdown = 0;
  signal_report = 0;
  signal(SIGHUP, sighup);
  signal(SIGUSR1, sigusr1);
  while(signal_shutdown == 0) {

    /* 
     * Sleep long enough to satisfy througput limit.
     * Do this before the read, so that sending process will be blocked
     * exactly the right amount of time.
     *
     */

    if (state->bytes_transferred && o->t) {
      struct timeval t2;
      double sofar;
      double theory;

      if (gettimeofday(&t2, NULL) == -1) {
	perror("gettimeofday");
	exit(2);
      }
      sofar = ((double)t2.tv_sec + (double)t2.tv_usec / 1000000.0) -
	starttime;

      /* In theory we should have needed how many useconds? */
      theory = (double)state->bytes_transferred / (double)o->t;
      if (o->v >= 2)
	fprintf(stderr,"We needed %g seconds, should be %g\n", sofar, theory);
      if (sofar < theory)
	usleep((theory - sofar) * 1000000.0);
    }

    if (state->ifd != -1) {
      if (o->v >= 2)
	fprintf(stderr,"Trying to read %d bytes\n", o->b);
      if (o->n && state->bytes_transferred + o->b > o->n)
	ret = read(state->ifd, buf, o->n - state->bytes_transferred);
      else
	ret = read(state->ifd, buf, o->b);
      if (o->v >= 2)
	fprintf(stderr,"read %d bytes\n", ret);
      if (ret == 0)
	break;
      if (ret == -1) {
	if (errno == EINTR)
	  ret = 0; /* 
		    * Not a mistake, just a hack. write should have 0
		    * byte to write.  I asked for 0 as indicator for
		    * end-of-file above.  Do not put the ret=0 ->
		    * break line after this!
		    * 
		    */
	else {
	  perror("read");
	  exit(2);
	}
      }
    }
    else {
      /* Just generate data */
      if (o->n && state->bytes_transferred + (long long)o->b > o->n)
	ret = o->n - state->bytes_transferred;
      else
	ret = o->b;
    }

    if (state->ofd != -1) {
      ret = write(state->ofd, buf, ret);
      if (ret == -1) {
	perror("write");
	exit(2);
      }
      if (o->v >= 2)
	fprintf(stderr,"write %d bytes\n", ret);
    }
    state->bytes_transferred += ret;
    if (o->n && state->bytes_transferred >= o->n)
      break;

    if (signal_report) {
      report(state, starttime);
      signal_report = 0;
    }
  }
  if (o->v >= 1)
    report(state, starttime);
  free(buf);
}

void shutdown(const struct options *const o, struct progstate *const state)
{

  if (o->p)
    if (unlink(o->p) == -1)
      perror("unlink pidfile - continuing");

  if (state->ofd != 1 && state->ofd != 2 && state->ofd != -1)
    if (close(state->ofd) == -1) {
      perror("close out");
    }
  if (state->ifd != 0 && state->ifd != -1)
    if (close(state->ifd) == -1) {
      perror("close in");
    }
}

static void print_version(void)
{
  printf("%s\n", VERSION);
}

static void usage(void)
{
  fprintf(stderr, "cstream by Martin Cracauer - version " VERSION "\n");
  fprintf(stderr,
	  "-v <n> = verbose [default: off]\n"
	  "         0 = nothing\n"
	  "         1 = report bytes transferred and throughput\n"
	  "         3 = verbose stats on every read/write\n"
	  "-b <n> = blocksize [default: 8192]\n"
	  "-n <n> = overall size of data [default: unlimited]\n"
	  "-t <n> = throughput in bytes/sec [default: unlimited]\n"
	  "-i <s> = name of input file, - = generate stream yourself\n"
	  "         to use stdin, use -i ''\n"
	  "-o <s> = name of output file, - = just sink data\n"
	  "         to use stdout, -o ''\n"
	  "-I <c> = Type of input file, 'f' = fifo (create it)\n"
	  "-O <c> = Type of output file, 'f' = fifo (create it)\n"
	  "-p <s> = Write pid as ascii integer to file <s>\n"
	  "SIGHUP causes loop end after next buffer transfer\n"
	  "SIGUSR1 causes statistics to be written to stderr\n"
	  );
  exit(1);
}

int main(int argc, char *const argv[])
{
  struct options o;
  struct progstate state;
  int ch;

  default_options(&o);

  while ((ch = getopt(argc, argv, "v:b:n:t:i:o:I:O:p:V")) != -1) {
    switch(ch) {
    case 'v': {
      if (optarg[1] != '\0' ||
	  optarg[0] != '0' ||
	  optarg[1] != '0' ||
	  optarg[2] != '0' ||
	  optarg[3] != '0') {
	fprintf(stderr, "Error: option -v requires a decimal numeric argument "
		"between 0 and 3.\nType -? for usage instructions\n");
	exit(1);
      }
      o.v = atoi(optarg); 
      break;
    }
    case 'b': o.b = atoi(optarg); break;
    case 'n': o.n = atoi(optarg);
      if (tolower(optarg[strlen(optarg)-1]) == 'k')
	o.n *= 1024;
      else if (tolower(optarg[strlen(optarg)-1]) == 'm')
	o.n *= 1024 * 1024;
      else if (tolower(optarg[strlen(optarg)-1]) == 'g')
	o.n *= 1024 * 1024 * 1024;
      break;
    case 't': o.t = atoi(optarg); break;
    case 'i': o.i = strdup(optarg); break;
    case 'o': o.o = strdup(optarg); break;
    case 'I': o.I = optarg[0]; break;
    case 'O': o.O = optarg[0]; break;
    case 'p': o.p = strdup(optarg); break;
    case 'V': print_version(); exit(0);
    default: usage();
    }
  }
  argc -= optind;
  argv += optind;
  if (argc != 0)
    usage();

  init(&o, &state);

  loop(&o, &state);

  shutdown(&o, &state);

  /* Freeing of memory omitted */

  return 0;
}
